#include <windows.h>
#include <stdint.h>
#include <stdio.h>
#include <xinput.h>
#include <dsound.h>
#include <math.h>

#define internal static
#define local_persist static
#define global_variable static

#define PI32 3.1415926535897932384

typedef uint8_t uint8;
typedef uint16_t uint16;
typedef uint32_t uint32;
typedef uint64_t uint64;

typedef int8_t int8;
typedef int16_t int16;
typedef int32_t int32;
typedef int64_t int64;

typedef int32 bool32;

typedef float real32;
typedef double real64;

struct Win32OffscreenBuffer {
	
	BITMAPINFO info;
	void * memory;
	int width;
	int height;
	int pitch;
};

// TODO @simon: this might not stay has a global.
global_variable bool32 gRunning = false;
global_variable Win32OffscreenBuffer gBackBuffer;
global_variable LPDIRECTSOUNDBUFFER gSecondaryBuffer;

struct Win32WindowDimension {
	int width;
	int height;
};

internal Win32WindowDimension win32GetWindowDimension( HWND window ) {

	RECT clientRect;
	GetClientRect( window, &clientRect );
	Win32WindowDimension windowDimension;
	windowDimension.width = clientRect.right - clientRect.left;
	windowDimension.height = clientRect.bottom - clientRect.top;

	return windowDimension;
}

#define X_INPUT_GET_STATE(name) DWORD WINAPI name(DWORD dwUserIndex, XINPUT_STATE *pState )
#define X_INPUT_SET_STATE(name) DWORD WINAPI name(DWORD dwUserIndex, XINPUT_VIBRATION * pVibration )
typedef X_INPUT_GET_STATE( x_input_get_state );
typedef X_INPUT_SET_STATE( x_input_set_state );

X_INPUT_GET_STATE( XInputGetStateStub ) {
	return ERROR_DEVICE_NOT_CONNECTED;
}

X_INPUT_SET_STATE( XInputSetStateStub ) {
	return ERROR_DEVICE_NOT_CONNECTED;
}

global_variable x_input_get_state * XInputGetState_ = XInputGetStateStub;
global_variable x_input_set_state * XInputSetState_ = XInputSetStateStub;
#define XInputGetState XInputGetState_
#define XInputSetState XInputSetState_

internal void win32LoadXInput( void ) {

	// TODO @simon: test on windows 8.
	HMODULE XInputLibrary = LoadLibrary( "xinput1_4.dll" );
	
	if ( XInputLibrary == NULL ) {
		XInputLibrary = LoadLibrary( "xinput9_1_0.dll" );		
	}
	
	if ( XInputLibrary == NULL ) {
		XInputLibrary = LoadLibrary( "xinput1_3.dll" );
	}

	if ( XInputLibrary != NULL ) {
		XInputGetState = ( x_input_get_state * ) GetProcAddress( XInputLibrary, "XInputGetState" );
		XInputSetState = ( x_input_set_state * ) GetProcAddress( XInputLibrary, "XInputSetState" );
	} else {
		// TODO @simon: log error.
	}
}

#define DIRECT_SOUND_CREATE(name) HRESULT WINAPI name( LPCGUID pcGuidDevice, LPDIRECTSOUND *ppDS, LPUNKNOWN pUnkOuter )
typedef DIRECT_SOUND_CREATE(direct_sound_create);

internal void win32InitializeDirectSound( HWND window, int32 samplePerSecond, int32 bufferSize ) {
	// Load the library.
	HMODULE DSoundLibrary = LoadLibrary( "dsound.dll" );
	
	if ( DSoundLibrary != NULL ) {

		direct_sound_create * DirectSoundCreate = ( direct_sound_create * ) GetProcAddress( DSoundLibrary, "DirectSoundCreate" );
		
		// TODO @simon: check this on other windows.
		LPDIRECTSOUND directSound;
		
		if ( DirectSoundCreate != NULL && SUCCEEDED( DirectSoundCreate( 0, &directSound, 0 ) ) ) {
			
			WAVEFORMATEX waveFormat = { 0 };
			waveFormat.wFormatTag = WAVE_FORMAT_PCM;
			waveFormat.nChannels = 2;
			waveFormat.nSamplesPerSec = samplePerSecond;
			waveFormat.wBitsPerSample = 16;
			waveFormat.nBlockAlign = ( waveFormat.nChannels * waveFormat.wBitsPerSample ) / 8;
			waveFormat.nAvgBytesPerSec = waveFormat.nSamplesPerSec * waveFormat.nBlockAlign;
			waveFormat.cbSize = 0;
			
			if ( SUCCEEDED( directSound->SetCooperativeLevel( window, DSSCL_PRIORITY ) ) ) {
			
				DSBUFFERDESC bufferDescription = { 0 };
				bufferDescription.dwSize = sizeof( bufferDescription );
				bufferDescription.dwFlags = DSBCAPS_PRIMARYBUFFER;
				
				LPDIRECTSOUNDBUFFER primaryBuffer;
				
				if ( SUCCEEDED( directSound->CreateSoundBuffer( &bufferDescription, &primaryBuffer, 0 ) ) ) {
				
					if ( SUCCEEDED( primaryBuffer->SetFormat( &waveFormat ) ) ) {
					} else {
						// TODO @simon: log error.
					}
					// bufferDescription.dwBufferBytes = 
				
				} else {
					// TODO @simon: log error.
				}
				
			} else {
				// TODO @simon: log error.
			}
			
			DSBUFFERDESC bufferDescription = { 0 };
			bufferDescription.dwSize = sizeof( bufferDescription );
			bufferDescription.dwFlags = 0;
			bufferDescription.dwBufferBytes = bufferSize;
			bufferDescription.lpwfxFormat = &waveFormat;

			if ( SUCCEEDED( directSound->CreateSoundBuffer( &bufferDescription, &gSecondaryBuffer, 0 ) ) ) {

			} else {
				// TODO @simon: log error.
			}
				
		} else {
			// TODO @simon: log error.
		}
	}
}

internal void renderWeirdGradient( Win32OffscreenBuffer *buffer, int xOffset, int yOffset ) {

	uint8 * row = ( uint8 * ) buffer->memory;

	for ( int y = 0; y < buffer->height; y++ ) {
	
		uint32 *pixel = ( uint32 *) row;

		for ( int x = 0; x < buffer->width; x++ ) {

			/*					 +3 +2 +1 +0 
				Pixel in memory: 00 00 00 00
								 aa rr gg bb 
			*/
			uint8 a = 0;
			uint8 r = ( x + xOffset );
			uint8 g = ( y + yOffset );
			uint8 b = 0;

			* pixel++ = a << 24 | r << 16 | g << 8 | b; 
		}

		row += buffer->pitch;
	}
}

internal void win32ResizeDIBSection( Win32OffscreenBuffer * buffer, int width, int height ) {

	if ( buffer->memory != 0 ) {
		VirtualFree( buffer->memory, 0, MEM_RELEASE );
	}

	buffer->width = width;
	buffer->height = height;
	int bytesPerPixel = 4;
	buffer->pitch = bytesPerPixel * buffer->width;

	buffer->info.bmiHeader.biSize = sizeof( buffer->info.bmiHeader );
	buffer->info.bmiHeader.biWidth = buffer->width;
	buffer->info.bmiHeader.biHeight = -buffer->height;
	buffer->info.bmiHeader.biPlanes = 1;
	buffer->info.bmiHeader.biBitCount = 32;
	buffer->info.bmiHeader.biCompression = BI_RGB;

	int bitmapMemorySize = ( buffer->width * buffer->height ) * bytesPerPixel; 
	buffer->memory = VirtualAlloc( 0, bitmapMemorySize, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE );
	// TODO @simon: initialize bitmapMemory to black;
}

internal void win32DisplayBufferInWindow( HDC deviceContext, int windowWidth, int windowHeight, Win32OffscreenBuffer * buffer ) {

	// TODO @simon: aspect ratio correction.
	StretchDIBits(
		deviceContext,
		0, 0, windowWidth, windowHeight,
		0, 0, buffer->width, buffer->height,
		buffer->memory,
		&buffer->info,
		DIB_RGB_COLORS,
		SRCCOPY
	);	
}

struct Win32SoundOutput {

    int samplesPerSecond;
	int frequency;
	int volume;
	uint32 runningSampleIndex;
	int wavePeriod;
	int bytesPerSample;
	int secondaryBufferSize;
	real32 tSine;
	int latencySampleCount;
};

internal void win32FillSoundBuffer( Win32SoundOutput * soundOutput, DWORD bytesToLock, DWORD bytesToWrite ) {

	VOID * region1;
	DWORD region1Size;
	VOID * region2;
	DWORD region2Size;
						
	if ( SUCCEEDED( gSecondaryBuffer->Lock(
		bytesToLock,
		bytesToWrite,
		&region1, &region1Size,
		&region2, &region2Size,
		0
		) ) ) {
	
		// TODO @simon: assert that regionXSize is valid.
		DWORD region1SampleCount = region1Size / soundOutput->bytesPerSample;
		int16 * sampleOut = ( int16 * ) region1;

		for ( DWORD sampleIndex = 0; sampleIndex < region1SampleCount; sampleIndex++ ) {
			
			//real32 t = 2.0f * PI32 * ( ( real32 ) soundOutput->runningSampleIndex / ( real32 ) soundOutput->wavePeriod );
			real32 sinValue = sinf( soundOutput->tSine );
			int16 sampleValue = ( int16 )( sinValue * soundOutput->volume );
			*sampleOut++ = sampleValue;
			*sampleOut++ = sampleValue;
			
			soundOutput->tSine += 2.0f * PI32 * 1.0f / ( real32 ) soundOutput->wavePeriod;
			soundOutput->runningSampleIndex++;
		}

		DWORD region2SampleCount = region2Size / soundOutput->bytesPerSample;
		sampleOut = ( int16 * ) region2;

		for ( DWORD sampleIndex = 0; sampleIndex < region2SampleCount; sampleIndex++ ) {

			//real32 t = 2.0f * PI32 * ( ( real32 ) soundOutput->runningSampleIndex / ( real32 ) soundOutput->wavePeriod );
			real32 sinValue = sinf( soundOutput->tSine );
			int16 sampleValue = ( int16 )( sinValue * soundOutput->volume );
			*sampleOut++ = sampleValue;
			*sampleOut++ = sampleValue;
			
			soundOutput->tSine += 2.0f * PI32 * 1.0f / ( real32 ) soundOutput->wavePeriod;
			soundOutput->runningSampleIndex++;
		}
		
		gSecondaryBuffer->Unlock( region1, region1Size, region2, region2Size );
	}
}

LRESULT CALLBACK windowProc(
	HWND window,
	UINT message,
	WPARAM wParam,
	LPARAM lParam
) {

	LRESULT result = 0;

	switch( message ) {

		case WM_SIZE: { 
			// Win32WindowDimension dimension = win32GetWindowDimension( window );
			// win32ResizeDIBSection( &gBackBuffer, dimension.width, dimension.height );
		} break;

		case WM_DESTROY: {
			// TODO @simon: this is an error, must recreate the window.
			gRunning = false;
		} break;

		case WM_CLOSE: {
			// TODO @simon: message to the user.
			gRunning = false;
		} break;

		case WM_ACTIVATEAPP: {
			OutputDebugStringA( "WM_ACTIVATEAPP\n" );
		} break;

		case WM_SYSKEYDOWN:
		case WM_SYSKEYUP:
		case WM_KEYDOWN:
		case WM_KEYUP: {

			uint32 VKCode = wParam;
			bool32 wasDown = lParam & ( 1 << 30 );
			bool32 isDown = lParam & ( 1 << 31 );

			if ( VKCode == 'W' ) {			
			}

			if ( VKCode == 'Z' ) {
			}

			if ( VKCode == 'A' ) {
			}

			if ( VKCode == 'Q' ) {
			}

			if ( VKCode == 'S' ) {
			}

			if ( VKCode == 'D' ) {
			}

			if ( VKCode == 'E' ) {
			}

			if ( VKCode == VK_ESCAPE ) {
			}

			if ( VKCode == VK_SPACE ) {
			}
			
			bool32 altKeyIsDown = lParam & ( 1 << 29 );

			if ( VKCode == VK_F4  && altKeyIsDown ) {
				gRunning = false;
			}

		} break;


		case WM_PAINT: {

			PAINTSTRUCT paint;

			HDC deviceContext = BeginPaint( window, &paint );
			Win32WindowDimension dimension = win32GetWindowDimension( window );
			win32DisplayBufferInWindow( deviceContext, dimension.width, dimension.height, &gBackBuffer );
			EndPaint( window, &paint );

		} break;

		default: {
			result = DefWindowProc( window, message, wParam, lParam ); 
		} break;
	}

	return result;
}

int CALLBACK WinMain(
	HINSTANCE instance,
	HINSTANCE previousInstance,
	LPSTR commandLine,
	int showCode
) {

	LARGE_INTEGER queryFrequency;
	QueryPerformanceFrequency( &queryFrequency );
	int64 performanceFrequency = queryFrequency.QuadPart;

	win32LoadXInput( );
	WNDCLASS windowClass = { 0 };

	windowClass.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
	windowClass.lpfnWndProc = windowProc; 
	windowClass.hInstance = instance; 
	//WindowClass.hIcon; 
	windowClass.lpszClassName = "HandmadeHeroWindowClass";
	
	if ( RegisterClass( &windowClass ) ) {

		HWND window = 
			CreateWindowEx(
				0,
				windowClass.lpszClassName,
				"Handmade Hero",
				WS_OVERLAPPEDWINDOW | WS_VISIBLE,
				CW_USEDEFAULT,
				CW_USEDEFAULT,
				CW_USEDEFAULT,
				CW_USEDEFAULT,
				0,
				0,
				instance,
				0	
			);

		if ( window != NULL ) {
			
			HDC deviceContext = GetDC( window );
			
			gRunning = true;
			int xOffset = 0;
			int yOffset = 0;
			
            Win32SoundOutput soundOutput = { 0 };

			// DirectSound variables.
			soundOutput.samplesPerSecond = 48000;
			soundOutput.frequency = 440;
			soundOutput.volume = 800;
			soundOutput.runningSampleIndex = 0;
			soundOutput.wavePeriod = soundOutput.samplesPerSecond / soundOutput.frequency;
			soundOutput.bytesPerSample = sizeof( int16 ) * 2;
			soundOutput.secondaryBufferSize = soundOutput.samplesPerSecond * soundOutput.bytesPerSample;
			soundOutput.tSine = 0;
			soundOutput.latencySampleCount = soundOutput.samplesPerSecond / 15;

			win32InitializeDirectSound( window, soundOutput.samplesPerSecond, soundOutput.secondaryBufferSize );
			win32FillSoundBuffer( &soundOutput, 0, soundOutput.latencySampleCount * soundOutput.bytesPerSample );
			gSecondaryBuffer->Play( 0, 0, DSBPLAY_LOOPING );
			win32ResizeDIBSection( &gBackBuffer, 1280, 720 );

			LARGE_INTEGER previousCounter;
			QueryPerformanceCounter( &previousCounter );
			
			uint64 previousCycleCounter = __rdtsc( );
			
			while ( gRunning ) {
				
				MSG message;

				while ( PeekMessage( &message, 0, 0, 0, PM_REMOVE ) ) {
					
					if ( message.message == WM_QUIT ) {
						gRunning = false;
					}
					
					TranslateMessage( &message );	
					DispatchMessage( &message );
				}

				// TODO @simon: should we poll this more frequently.
				for ( DWORD controllerIndex = 0; controllerIndex < XUSER_MAX_COUNT; controllerIndex++ ) {

					XINPUT_STATE controllerState;

					if ( XInputGetState( controllerIndex, &controllerState ) == ERROR_SUCCESS ) {
						// Controller plugged in.
						// TODO @simon: check dwPacketNumber
						XINPUT_GAMEPAD *gamepad = &controllerState.Gamepad;

						bool32 up = ( gamepad->wButtons & XINPUT_GAMEPAD_DPAD_UP );
						bool32 down = ( gamepad->wButtons & XINPUT_GAMEPAD_DPAD_DOWN );
						bool32 left = ( gamepad->wButtons & XINPUT_GAMEPAD_DPAD_LEFT );
						bool32 right = ( gamepad->wButtons & XINPUT_GAMEPAD_DPAD_RIGHT );
						bool32 start = ( gamepad->wButtons & XINPUT_GAMEPAD_START );
						bool32 back = ( gamepad->wButtons & XINPUT_GAMEPAD_BACK );
						bool32 leftShoulder = ( gamepad->wButtons & XINPUT_GAMEPAD_LEFT_SHOULDER );
						bool32 rightShoulder = ( gamepad->wButtons & XINPUT_GAMEPAD_RIGHT_SHOULDER );
						bool32 AButton = ( gamepad->wButtons & XINPUT_GAMEPAD_A );
						bool32 BButton = ( gamepad->wButtons & XINPUT_GAMEPAD_B );
						bool32 XButton = ( gamepad->wButtons & XINPUT_GAMEPAD_X );
						bool32 YButton = ( gamepad->wButtons & XINPUT_GAMEPAD_Y );
						
						int16 stickX = gamepad->sThumbLX;
						int16 stickY = gamepad->sThumbLY;

						xOffset += stickX / 4096;
						yOffset += stickY / 4096;

						soundOutput.frequency = 440 + ( int ) ( 512.0f * ( ( real32 ) stickY / 32767.0f ) );
						soundOutput.wavePeriod = soundOutput.samplesPerSecond / soundOutput.frequency;
						
					} else {
						// Controller not available.
					}
				}

				/*XINPUT_VIBRATION vibration;
				vibration.wLeftMotorSpeed = 6000;
				vibration.wRightMotorSpeed = 5000;
				XInputSetState( 0, &vibration );*/

				renderWeirdGradient( &gBackBuffer, xOffset, yOffset );
				
				// NOTE @simon: DirectSound output test.
				
				DWORD playCursor;
				DWORD writeCursor;
				
				if ( SUCCEEDED( gSecondaryBuffer->GetCurrentPosition( &playCursor, &writeCursor ) ) ) {
				
					DWORD bytesToLock = ( soundOutput.runningSampleIndex * soundOutput.bytesPerSample ) % soundOutput.secondaryBufferSize;
					DWORD targetCursor = ( playCursor + ( soundOutput.latencySampleCount * soundOutput.bytesPerSample ) ) % soundOutput.secondaryBufferSize;
					DWORD bytesToWrite;
					
					// TODO @simon: change this to a lower latency.
					if ( bytesToLock > targetCursor ) {
						bytesToWrite = soundOutput.secondaryBufferSize - bytesToLock;
						bytesToWrite += targetCursor;
					} else {
						bytesToWrite = targetCursor - bytesToLock;
					}

					win32FillSoundBuffer( &soundOutput, bytesToLock, bytesToWrite );
				}

				Win32WindowDimension dimension = win32GetWindowDimension( window );
				win32DisplayBufferInWindow( deviceContext, dimension.width, dimension.height, &gBackBuffer );
				
				uint64 endCycleCounter = __rdtsc( );
				uint64 elapsedCycles = endCycleCounter - previousCycleCounter;
				real32 megaCycles = elapsedCycles / 1000000.0f;
				previousCycleCounter = endCycleCounter;
				
				LARGE_INTEGER endCounter;
				QueryPerformanceCounter( &endCounter );
				int64 counterDifference = endCounter.QuadPart - previousCounter.QuadPart;
				real32 msPerFrame = ( ( 1000.0f * counterDifference ) / performanceFrequency );
				real32 framePerSecond = ( real32 ) performanceFrequency / counterDifference;
				
				char buffer[ 256 ];
				sprintf( buffer, "Frame time: %.04f ms / %.04f fps / %.04f Mc/f\n", msPerFrame, framePerSecond, megaCycles );
                OutputDebugString( buffer );
				
				previousCounter = endCounter;
			}

			ReleaseDC( window, deviceContext );
			
		} else {
			// TODO @simon: logging
		}

	} else {
		// TODO @simon: logging
	}

	return 0;
}
